/*
 * Written by Alexei V. Galatenko <agalat@castle.nmd.msu.ru>
 */

#include <pniam.h>

char *get_pniam_err (pniam_result_t res)
{
    switch (res)
    {
        case PNIAM_SYSERR:
            return "SYSERR";
        case PNIAM_FAIL:
            return "PNIAM_FAIL";
        case PNIAM_AUTHTOKERR:
            return "Authentication token error";
        break;
    }
    return "Error";
}
