int utmp_open_session(pam_handle_t *pamh, pid_t pid
                             , const char **callname, const char **err_descr);
int utmp_close_session(pam_handle_t *pamh
                              , const char **callname, const char **err_descr);
