#include <pniam.h>

int utmp_open_session(pniam_item_list_t *list, pid_t pid
                             , const char **callname, const char **err_descr);
int utmp_close_session(const char **callname, const char **err_descr);
